@@ -1,7 +1,7 @@
 /******************************************************************************
  *
  * Project:  PROJ
- * Purpose:  ISO19111:2018 implementation
+ * Purpose:  ISO19111:2019 implementation
  * Author:   Even Rouault <even dot rouault at spatialys dot com>
  *
  ******************************************************************************
@@ -63,15 +63,15 @@ namespace operation {
 
 /** \brief Grid description */
 struct GridDescription {
-    std::string shortName{};   /**< Grid short filename */
-    std::string fullName{};    /**< Grid full path name (if found) */
-    std::string packageName{}; /**< Package name (or empty) */
-    std::string url{}; /**< Grid URL (if packageName is empty), or package
-                            URL (or empty) */
-    bool directDownload = false; /**< Whether url can be fetched directly. */
-    bool openLicense =
-        false; /**< Whether the grid is released with an open license. */
-    bool available = false; /**< Whether GRID is available. */
+    std::string shortName;   /**< Grid short filename */
+    std::string fullName;    /**< Grid full path name (if found) */
+    std::string packageName; /**< Package name (or empty) */
+    std::string url;         /**< Grid URL (if packageName is empty), or package
+                                    URL (or empty) */
+    bool directDownload;     /**< Whether url can be fetched directly. */
+    /** Whether the grid is released with an open license. */
+    bool openLicense;
+    bool available; /**< Whether GRID is available. */
 
     //! @cond Doxygen_Suppress
     bool operator<(const GridDescription &other) const {
@@ -113,10 +113,11 @@ using CoordinateOperationNNPtr = util::nn<CoordinateOperationPtr>;
  * (some) entirely different parameter values are needed, a different coordinate
  * operation shall be defined.
  *
- * \remark Implements CoordinateOperation from \ref ISO_19111_2018
+ * \remark Implements CoordinateOperation from \ref ISO_19111_2019
  */
 class PROJ_GCC_DLL CoordinateOperation : public common::ObjectUsage,
-                                         public io::IPROJStringExportable {
+                                         public io::IPROJStringExportable,
+                                         public io::IJSONExportable {
   public:
     //! @cond Doxygen_Suppress
     PROJ_DLL ~CoordinateOperation() override;
@@ -143,13 +144,25 @@ class PROJ_GCC_DLL CoordinateOperation : public common::ObjectUsage,
 
     /** \brief Return grids needed by an operation. */
     PROJ_DLL virtual std::set<GridDescription>
-    gridsNeeded(const io::DatabaseContextPtr &databaseContext) const = 0;
+    gridsNeeded(const io::DatabaseContextPtr &databaseContext,
+                bool considerKnownGridsAsAvailable) const = 0;
 
     PROJ_DLL bool
-    isPROJInstantiable(const io::DatabaseContextPtr &databaseContext) const;
+    isPROJInstantiable(const io::DatabaseContextPtr &databaseContext,
+                       bool considerKnownGridsAsAvailable) const;
 
     PROJ_DLL bool hasBallparkTransformation() const;
 
+    PROJ_DLL static const std::string OPERATION_VERSION_KEY;
+
+    PROJ_DLL CoordinateOperationNNPtr normalizeForVisualization() const;
+
+    PROJ_PRIVATE :
+        //! @cond Doxygen_Suppress
+        PROJ_FOR_TEST CoordinateOperationNNPtr
+        shallowClone() const;
+    //! @endcond
+
   protected:
     PROJ_INTERNAL CoordinateOperation();
     PROJ_INTERNAL CoordinateOperation(const CoordinateOperation &other);
@@ -171,6 +184,12 @@ class PROJ_GCC_DLL CoordinateOperation : public common::ObjectUsage,
         const std::vector<metadata::PositionalAccuracyNNPtr> &accuracies);
     PROJ_INTERNAL void setHasBallparkTransformation(bool b);
 
+    PROJ_INTERNAL void
+    setProperties(const util::PropertyMap
+                      &properties); // throw(InvalidValueTypeException)
+
+    PROJ_INTERNAL virtual CoordinateOperationNNPtr _shallowClone() const = 0;
+
   private:
     PROJ_OPAQUE_PRIVATE_DATA
     CoordinateOperation &operator=(const CoordinateOperation &other) = delete;
@@ -181,7 +200,7 @@ class PROJ_GCC_DLL CoordinateOperation : public common::ObjectUsage,
 /** \brief Abstract class modelling a parameter value (OperationParameter)
  * or group of parameters.
  *
- * \remark Implements GeneralOperationParameter from \ref ISO_19111_2018
+ * \remark Implements GeneralOperationParameter from \ref ISO_19111_2019
  */
 class PROJ_GCC_DLL GeneralOperationParameter : public common::IdentifiedObject {
   public:
@@ -193,7 +212,8 @@ class PROJ_GCC_DLL GeneralOperationParameter : public common::IdentifiedObject {
     PROJ_INTERNAL bool _isEquivalentTo(
         const util::IComparable *other,
         util::IComparable::Criterion criterion =
-            util::IComparable::Criterion::STRICT) const override = 0;
+            util::IComparable::Criterion::STRICT,
+        const io::DatabaseContextPtr &dbContext = nullptr) const override = 0;
     //! @endcond
 
   protected:
@@ -225,7 +245,7 @@ using OperationParameterNNPtr = util::nn<OperationParameterPtr>;
  * Most parameter values are numeric, but other types of parameter values are
  * possible.
  *
- * \remark Implements OperationParameter from \ref ISO_19111_2018
+ * \remark Implements OperationParameter from \ref ISO_19111_2019
  */
 class PROJ_GCC_DLL OperationParameter final : public GeneralOperationParameter {
   public:
@@ -234,10 +254,11 @@ class PROJ_GCC_DLL OperationParameter final : public GeneralOperationParameter {
     //! @endcond
 
     //! @cond Doxygen_Suppress
-    PROJ_INTERNAL bool
-    _isEquivalentTo(const util::IComparable *other,
-                    util::IComparable::Criterion criterion =
-                        util::IComparable::Criterion::STRICT) const override;
+    PROJ_INTERNAL bool _isEquivalentTo(
+        const util::IComparable *other,
+        util::IComparable::Criterion criterion =
+            util::IComparable::Criterion::STRICT,
+        const io::DatabaseContextPtr &dbContext = nullptr) const override;
     //! @endcond
 
     // non-standard
@@ -271,10 +292,11 @@ struct MethodMapping;
 /** \brief Abstract class modelling a parameter value (OperationParameterValue)
  * or group of parameter values.
  *
- * \remark Implements GeneralParameterValue from \ref ISO_19111_2018
+ * \remark Implements GeneralParameterValue from \ref ISO_19111_2019
  */
 class PROJ_GCC_DLL GeneralParameterValue : public util::BaseObject,
                                            public io::IWKTExportable,
+                                           public io::IJSONExportable,
                                            public util::IComparable {
   public:
     //! @cond Doxygen_Suppress
@@ -283,10 +305,14 @@ class PROJ_GCC_DLL GeneralParameterValue : public util::BaseObject,
     PROJ_INTERNAL void _exportToWKT(io::WKTFormatter *formatter)
         const override = 0; // throw(io::FormattingException)
 
+    PROJ_INTERNAL void _exportToJSON(io::JSONFormatter *formatter)
+        const override = 0; // throw(FormattingException)
+
     PROJ_INTERNAL bool _isEquivalentTo(
         const util::IComparable *other,
         util::IComparable::Criterion criterion =
-            util::IComparable::Criterion::STRICT) const override = 0;
+            util::IComparable::Criterion::STRICT,
+        const io::DatabaseContextPtr &dbContext = nullptr) const override = 0;
     //! @endcond
 
   protected:
@@ -325,7 +351,7 @@ using ParameterValueNNPtr = util::nn<ParameterValuePtr>;
  * Most parameter values are numeric, but other types of parameter values are
  * possible.
  *
- * \remark Implements ParameterValue from \ref ISO_19111_2018
+ * \remark Implements ParameterValue from \ref ISO_19111_2019
  */
 class PROJ_GCC_DLL ParameterValue final : public util::BaseObject,
                                           public io::IWKTExportable,
@@ -369,10 +395,11 @@ class PROJ_GCC_DLL ParameterValue final : public util::BaseObject,
     PROJ_DLL bool booleanValue() PROJ_PURE_DECL;
 
     //! @cond Doxygen_Suppress
-    PROJ_INTERNAL bool
-    _isEquivalentTo(const util::IComparable *other,
-                    util::IComparable::Criterion criterion =
-                        util::IComparable::Criterion::STRICT) const override;
+    PROJ_INTERNAL bool _isEquivalentTo(
+        const util::IComparable *other,
+        util::IComparable::Criterion criterion =
+            util::IComparable::Criterion::STRICT,
+        const io::DatabaseContextPtr &dbContext = nullptr) const override;
     //! @endcond
 
   protected:
@@ -400,7 +427,7 @@ using OperationParameterValueNNPtr = util::nn<OperationParameterValuePtr>;
  *
  * This combines a OperationParameter with the corresponding ParameterValue.
  *
- * \remark Implements OperationParameterValue from \ref ISO_19111_2018
+ * \remark Implements OperationParameterValue from \ref ISO_19111_2019
  */
 class PROJ_GCC_DLL OperationParameterValue final
     : public GeneralParameterValue {
@@ -426,10 +453,14 @@ class PROJ_GCC_DLL OperationParameterValue final
     PROJ_INTERNAL void _exportToWKT(io::WKTFormatter *formatter)
         const override; // throw(io::FormattingException)
 
-    PROJ_INTERNAL bool
-    _isEquivalentTo(const util::IComparable *other,
-                    util::IComparable::Criterion criterion =
-                        util::IComparable::Criterion::STRICT) const override;
+    PROJ_INTERNAL void _exportToJSON(io::JSONFormatter *formatter)
+        const override; // throw(FormattingException)
+
+    PROJ_INTERNAL bool _isEquivalentTo(
+        const util::IComparable *other,
+        util::IComparable::Criterion criterion =
+            util::IComparable::Criterion::STRICT,
+        const io::DatabaseContextPtr &dbContext = nullptr) const override;
     //! @endcond
 
   protected:
@@ -463,9 +494,10 @@ using OperationMethodNNPtr = util::nn<OperationMethodPtr>;
  * For a projection method, this contains the name of the projection method
  * and the name of the projection parameters.
  *
- * \remark Implements OperationMethod from \ref ISO_19111_2018
+ * \remark Implements OperationMethod from \ref ISO_19111_2019
  */
-class PROJ_GCC_DLL OperationMethod : public common::IdentifiedObject {
+class PROJ_GCC_DLL OperationMethod : public common::IdentifiedObject,
+                                     public io::IJSONExportable {
   public:
     //! @cond Doxygen_Suppress
     PROJ_DLL ~OperationMethod() override;
@@ -491,10 +523,14 @@ class PROJ_GCC_DLL OperationMethod : public common::IdentifiedObject {
     PROJ_INTERNAL void _exportToWKT(io::WKTFormatter *formatter)
         const override; // throw(io::FormattingException)
 
-    PROJ_INTERNAL bool
-    _isEquivalentTo(const util::IComparable *other,
-                    util::IComparable::Criterion criterion =
-                        util::IComparable::Criterion::STRICT) const override;
+    PROJ_INTERNAL void _exportToJSON(io::JSONFormatter *formatter)
+        const override; // throw(FormattingException)
+
+    PROJ_INTERNAL bool _isEquivalentTo(
+        const util::IComparable *other,
+        util::IComparable::Criterion criterion =
+            util::IComparable::Criterion::STRICT,
+        const io::DatabaseContextPtr &dbContext = nullptr) const override;
     //! @endcond
 
   protected:
@@ -534,7 +570,7 @@ using SingleOperationNNPtr = util::nn<SingleOperationPtr>;
 /** \brief A single (not concatenated) coordinate operation
  * (CoordinateOperation)
  *
- * \remark Implements SingleOperation from \ref ISO_19111_2018
+ * \remark Implements SingleOperation from \ref ISO_19111_2019
  */
 class PROJ_GCC_DLL SingleOperation : virtual public CoordinateOperation {
   public:
@@ -567,7 +603,8 @@ class PROJ_GCC_DLL SingleOperation : virtual public CoordinateOperation {
             std::vector<metadata::PositionalAccuracyNNPtr>());
 
     PROJ_DLL std::set<GridDescription>
-    gridsNeeded(const io::DatabaseContextPtr &databaseContext) const override;
+    gridsNeeded(const io::DatabaseContextPtr &databaseContext,
+                bool considerKnownGridsAsAvailable) const override;
 
     PROJ_DLL std::list<std::string> validateParameters() const;
 
@@ -579,13 +616,19 @@ class PROJ_GCC_DLL SingleOperation : virtual public CoordinateOperation {
                               const common::UnitOfMeasure &targetUnit) const
         noexcept;
 
+    PROJ_INTERNAL double
+    parameterValueNumeric(const char *param_name,
+                          const common::UnitOfMeasure &targetUnit) const
+        noexcept;
+
     PROJ_INTERNAL double parameterValueNumericAsSI(int epsg_code) const
         noexcept;
 
-    PROJ_INTERNAL bool
-    _isEquivalentTo(const util::IComparable *other,
-                    util::IComparable::Criterion criterion =
-                        util::IComparable::Criterion::STRICT) const override;
+    PROJ_INTERNAL bool _isEquivalentTo(
+        const util::IComparable *other,
+        util::IComparable::Criterion criterion =
+            util::IComparable::Criterion::STRICT,
+        const io::DatabaseContextPtr &dbContext = nullptr) const override;
     //! @endcond
 
   protected:
@@ -604,6 +647,7 @@ class PROJ_GCC_DLL SingleOperation : virtual public CoordinateOperation {
 
     PROJ_INTERNAL bool _isEquivalentTo(const util::IComparable *other,
                                        util::IComparable::Criterion criterion,
+                                       const io::DatabaseContextPtr &dbContext,
                                        bool inOtherDirection) const;
 
   private:
@@ -633,7 +677,7 @@ using ConversionNNPtr = util::nn<ConversionPtr>;
  * associations, but through associations from crs::DerivedCRS to
  * crs::SingleCRS.
  *
- * \remark Implements Conversion from \ref ISO_19111_2018
+ * \remark Implements Conversion from \ref ISO_19111_2019
  */
 
 /*!
@@ -1282,10 +1326,28 @@ class PROJ_GCC_DLL Conversion : public SingleOperation {
                      const common::Length &falseEasting,
                      const common::Length &falseNorthing);
 
+    PROJ_DLL static ConversionNNPtr
+    createVerticalPerspective(const util::PropertyMap &properties,
+                              const common::Angle &topoOriginLat,
+                              const common::Angle &topoOriginLong,
+                              const common::Length &topoOriginHeight,
+                              const common::Length &viewPointHeight,
+                              const common::Length &falseEasting,
+                              const common::Length &falseNorthing);
+
+    PROJ_DLL static ConversionNNPtr createPoleRotationGRIBConvention(
+        const util::PropertyMap &properties,
+        const common::Angle &southPoleLatInUnrotatedCRS,
+        const common::Angle &southPoleLongInUnrotatedCRS,
+        const common::Angle &axisRotation);
+
     PROJ_DLL static ConversionNNPtr
     createChangeVerticalUnit(const util::PropertyMap &properties,
                              const common::Scale &factor);
 
+    PROJ_DLL static ConversionNNPtr
+    createHeightDepthReversal(const util::PropertyMap &properties);
+
     PROJ_DLL static ConversionNNPtr createAxisOrderReversal(bool is3D);
 
     PROJ_DLL static ConversionNNPtr
@@ -1299,6 +1361,9 @@ class PROJ_GCC_DLL Conversion : public SingleOperation {
         _exportToPROJString(io::PROJStringFormatter *formatter)
             const override; // throw(FormattingException)
 
+    PROJ_INTERNAL void _exportToJSON(io::JSONFormatter *formatter)
+        const override; // throw(FormattingException)
+
     PROJ_INTERNAL const char *getESRIMethodName() const;
 
     PROJ_INTERNAL const char *getWKT1GDALMethodName() const;
@@ -1308,6 +1373,10 @@ class PROJ_GCC_DLL Conversion : public SingleOperation {
     PROJ_INTERNAL ConversionNNPtr alterParametersLinearUnit(
         const common::UnitOfMeasure &unit, bool convertToNewUnit) const;
 
+    PROJ_INTERNAL static ConversionNNPtr
+    createGeographicGeocentric(const crs::CRSNNPtr &sourceCRS,
+                               const crs::CRSNNPtr &targetCRS);
+
     //! @endcond
 
   protected:
@@ -1320,6 +1389,8 @@ class PROJ_GCC_DLL Conversion : public SingleOperation {
     PROJ_FRIEND(crs::ProjectedCRS);
     PROJ_INTERNAL bool addWKTExtensionNode(io::WKTFormatter *formatter) const;
 
+    PROJ_INTERNAL CoordinateOperationNNPtr _shallowClone() const override;
+
   private:
     PROJ_OPAQUE_PRIVATE_DATA
     Conversion &operator=(const Conversion &other) = delete;
@@ -1353,7 +1424,7 @@ using TransformationNNPtr = util::nn<TransformationPtr>;
  * coordinates are referenced to different datums (reference frames) will be a
  * coordinate transformation.
  *
- * \remark Implements Transformation from \ref ISO_19111_2018
+ * \remark Implements Transformation from \ref ISO_19111_2019
  */
 class PROJ_GCC_DLL Transformation : public SingleOperation {
   public:
@@ -1458,7 +1529,8 @@ class PROJ_GCC_DLL Transformation : public SingleOperation {
     PROJ_DLL static TransformationNNPtr
     createGravityRelatedHeightToGeographic3D(
         const util::PropertyMap &properties, const crs::CRSNNPtr &sourceCRSIn,
-        const crs::CRSNNPtr &targetCRSIn, const std::string &filename,
+        const crs::CRSNNPtr &targetCRSIn, const crs::CRSPtr &interpolationCRSIn,
+        const std::string &filename,
         const std::vector<metadata::PositionalAccuracyNNPtr> &accuracies);
 
     PROJ_DLL static TransformationNNPtr createVERTCON(
@@ -1516,6 +1588,9 @@ class PROJ_GCC_DLL Transformation : public SingleOperation {
     PROJ_INTERNAL void _exportToWKT(io::WKTFormatter *formatter)
         const override; // throw(io::FormattingException)
 
+    PROJ_INTERNAL void _exportToJSON(io::JSONFormatter *formatter)
+        const override; // throw(FormattingException)
+
     PROJ_INTERNAL TransformationNNPtr shallowClone() const;
 
     //! @endcond
@@ -1533,8 +1608,11 @@ class PROJ_GCC_DLL Transformation : public SingleOperation {
     PROJ_INTERNAL void _exportToPROJString(io::PROJStringFormatter *formatter)
         const override; // throw(FormattingException)
 
+    PROJ_FRIEND(CoordinateOperationFactory);
     PROJ_INTERNAL TransformationNNPtr inverseAsTransformation() const;
 
+    PROJ_INTERNAL CoordinateOperationNNPtr _shallowClone() const override;
+
   private:
     PROJ_OPAQUE_PRIVATE_DATA
 };
@@ -1553,7 +1631,7 @@ using PointMotionOperationNNPtr = util::nn<PointMotionOperationPtr>;
  *
  * The motion is due to tectonic plate movement or deformation.
  *
- * \remark Implements PointMotionOperation from \ref ISO_19111_2018
+ * \remark Implements PointMotionOperation from \ref ISO_19111_2019
  */
 class PROJ_GCC_DLL PointMotionOperation : public SingleOperation {
   public:
@@ -1582,7 +1660,7 @@ using ConcatenatedOperationNNPtr = util::nn<ConcatenatedOperationPtr>;
  * the source coordinate reference system of step n+1 shall be the same as
  * the target coordinate reference system of step n.
  *
- * \remark Implements ConcatenatedOperation from \ref ISO_19111_2018
+ * \remark Implements ConcatenatedOperation from \ref ISO_19111_2019
  */
 class PROJ_GCC_DLL ConcatenatedOperation final : public CoordinateOperation {
   public:
@@ -1605,7 +1683,8 @@ class PROJ_GCC_DLL ConcatenatedOperation final : public CoordinateOperation {
         bool checkExtent); // throw InvalidOperation
 
     PROJ_DLL std::set<GridDescription>
-    gridsNeeded(const io::DatabaseContextPtr &databaseContext) const override;
+    gridsNeeded(const io::DatabaseContextPtr &databaseContext,
+                bool considerKnownGridsAsAvailable) const override;
 
     PROJ_PRIVATE :
 
@@ -1614,10 +1693,14 @@ class PROJ_GCC_DLL ConcatenatedOperation final : public CoordinateOperation {
         _exportToWKT(io::WKTFormatter *formatter)
             const override; // throw(io::FormattingException)
 
-    PROJ_INTERNAL bool
-    _isEquivalentTo(const util::IComparable *other,
-                    util::IComparable::Criterion criterion =
-                        util::IComparable::Criterion::STRICT) const override;
+    PROJ_INTERNAL bool _isEquivalentTo(
+        const util::IComparable *other,
+        util::IComparable::Criterion criterion =
+            util::IComparable::Criterion::STRICT,
+        const io::DatabaseContextPtr &dbContext = nullptr) const override;
+
+    PROJ_INTERNAL void _exportToJSON(io::JSONFormatter *formatter)
+        const override; // throw(FormattingException)
 
     PROJ_INTERNAL static void
     fixStepsDirection(const crs::CRSNNPtr &concatOpSourceCRS,
@@ -1626,12 +1709,15 @@ class PROJ_GCC_DLL ConcatenatedOperation final : public CoordinateOperation {
     //! @endcond
 
   protected:
+    PROJ_INTERNAL ConcatenatedOperation(const ConcatenatedOperation &other);
     PROJ_INTERNAL explicit ConcatenatedOperation(
         const std::vector<CoordinateOperationNNPtr> &operationsIn);
 
     PROJ_INTERNAL void _exportToPROJString(io::PROJStringFormatter *formatter)
         const override; // throw(FormattingException)
 
+    PROJ_INTERNAL CoordinateOperationNNPtr _shallowClone() const override;
+
     INLINED_MAKE_SHARED
 
   private:
@@ -1728,6 +1814,12 @@ class PROJ_GCC_DLL CoordinateOperationContext {
         /** Ignore grid availability at all. Results will be presented as if
          * all grids were available. */
         IGNORE_GRID_AVAILABILITY,
+
+        /** Results will be presented as if grids known to PROJ (that is
+        * registered in the grid_alternatives table of its database) were
+        * available. Used typically when networking is enabled.
+        */
+        KNOWN_AVAILABLE,
     };
 
     PROJ_DLL void setGridAvailabilityUse(GridAvailabilityUse use);
