@@ -339,7 +339,7 @@ static int isea_snyder_forward(struct isea_geo * ll, struct isea_pt * out)
     double          theta;
 
     /* additional variables from snyder */
-    double          q, H, Ag, Azprime, Az, dprime, f, rho,
+    double          q, Rprime, H, Ag, Azprime, Az, dprime, f, rho,
                     x, y;
 
     /* variables used to store intermediate results */
@@ -427,7 +427,7 @@ static int isea_snyder_forward(struct isea_geo * ll, struct isea_pt * out)
         /* eq 5 */
         /* Rprime = 0.9449322893 * R; */
         /* R' in the paper is for the truncated */
-        const double Rprime = 0.91038328153090290025;
+        Rprime = 0.91038328153090290025;
 
         /* eq 6 */
         H = acos(sin(Az) * sin(G) * cos(g) - cos(Az) * cos(G));
@@ -733,7 +733,7 @@ static int isea_dddi_ap3odd(struct isea_dgg *g, int quadz, struct isea_pt *pt,
             quadz += 5;
             d = 0;
         }
-    } else if (quadz >= 6) {
+    } else /* if (quadz >= 6) */ {
         if (i == 0 && d == maxcoord) {
             /* south pole */
             quadz = 11;
@@ -812,7 +812,7 @@ static int isea_dddi(struct isea_dgg *g, int quadz, struct isea_pt *pt,
             h.y = -h.z;
             h.x = 0;
         }
-    } else if (quadz >= 6) {
+    } else /* if (quadz >= 6) */ {
         if (h.z == 0 && h.x == sidelength) {
             /* south pole */
             quadz = 11;
