@@ -348,7 +348,9 @@ TEST(gie, info_functions) {
 
     /* proj_info()                                                            */
     /* this one is difficult to test, since the output changes with the setup */
+    putenv(const_cast<char *>("PROJ_SKIP_READ_USER_WRITABLE_DIRECTORY="));
     info = proj_info();
+    putenv(const_cast<char *>("PROJ_SKIP_READ_USER_WRITABLE_DIRECTORY=YES"));
 
     if (info.version[0] != '\0') {
         char tmpstr[64];
@@ -360,6 +362,9 @@ TEST(gie, info_functions) {
         ASSERT_NE(std::string(info.searchpath), std::string());
     }
 
+    ASSERT_TRUE(std::string(info.searchpath).find("/proj") !=
+                std::string::npos);
+
     /* proj_pj_info() */
     {
         P = proj_create(PJ_DEFAULT_CTX,
@@ -379,9 +384,10 @@ TEST(gie, info_functions) {
     proj_destroy(P);
 
     /* proj_grid_info() */
-    grid_info = proj_grid_info("null");
+    grid_info = proj_grid_info("tests/test_hgrid.tif");
     ASSERT_NE(std::string(grid_info.filename), "");
-    ASSERT_EQ(std::string(grid_info.gridname), "null");
+    ASSERT_EQ(std::string(grid_info.gridname), "tests/test_hgrid.tif");
+    ASSERT_EQ(std::string(grid_info.format), "gtiff");
 
     grid_info = proj_grid_info("nonexistinggrid");
     ASSERT_EQ(std::string(grid_info.filename), "");
